from __future__ import annotations
import asyncio
import sys

from typing import Any, Mapping
from quart import Quart

import yaml
from simpleconf import ProfileConfig
from simpleconf.utils import POOL_KEY

from .logger import get_logger
from .config import (
    CONFIG_SCHEMA,
    MY_CONFIG_FILE,
    config,
)
from .rings._callbacks import StreamingWebsocketCallbackHandler
from .manager import Manager


class WS:

    def __init__(self) -> None:
        self.config = config
        self.logger = get_logger(config.log_level)
        self.ws = None
        # Each ws connection should have one profile
        self._profile = "default"
        self.manager = None

    @property
    def profile(self) -> str:
        return self._profile

    @profile.setter
    def profile(self, value: str) -> None:
        if value == self.profile:
            return
        self._profile = value
        ProfileConfig.use_profile(self.config, self.profile)

    async def on_connect(
        self,
        app: Quart,
        data: Mapping[str, Any],
    ) -> None:
        """Run on websocket connect"""
        self.logger.info("Websocket connected from client")
        # determine profile, from the web.
        await self.ws.send_json({"event": "get_saved_profile"})

    async def on_saved_profile_obtained(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket saved profile obtained"""
        self.profile = data["profile"]
        # profile determined, init the manager
        self.manager = Manager(self.profile, serving=True)

        await self.ws.send_json(
            {
                "event": "initialized",
                "histories": self.manager.get_histories(True),
                "profiles": ProfileConfig.profiles(self.config),
                "ingestable": bool(self.config.ingest.source_directory),
            }
        )

    async def on_disconnect(self, app: Quart) -> None:
        """Run on websocket disconnect"""
        self.logger.info("Websocket disconnected from client")
        self.ws = None

    async def on_query(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket query"""
        if not self.manager:
            await self.ws.send_json(
                {
                    "event": "lc_error",
                    "error": "Manager not initialized",
                    "type": "custom",
                }
            )

        query, conversation = (
            data["query"],
            data.get("conversation", None),
        )
        if not conversation:
            title = query.splitlines()[0]
            conversation = self.manager.new_history(title=title)
            await self.ws.send_json(
                {
                    "event": "query_started",
                    "conversation": conversation,
                    "title": title,
                }
            )

        async def query_in_background() -> None:
            """Run query in background"""
            try:
                await self.manager.a_query(
                    query,
                    conversation,
                    StreamingWebsocketCallbackHandler(self.ws),
                )
            except Exception as ex:
                self.logger.exception(str(ex))
                await self.ws.send_json(
                    {
                        "event": "lc_error",
                        "error": str(ex),
                        "type": "custom",
                    }
                )

        app.add_background_task(query_in_background)

    async def on_load_conversation(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket load conversation"""
        conversation = data["conversation"]
        messages = self.manager.load_history(conversation).get("history", [])
        await self.ws.send_json(
            {
                "event": "conversation_loaded",
                "messages": [
                    {
                        "when": "Earlier",
                        "who": message["type"],
                        "message": message["data"]["content"],
                    }
                    for message in messages
                ],
            }
        )

    async def on_delete_conversation(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket delete conversation"""
        conversation = data["conversation"]
        try:
            self.manager.history_file(conversation).unlink()
        except Exception as ex:
            await self.ws.send_json(
                {
                    "event": "conversation_deleted",
                    "ok": False,
                    "error": str(ex),
                    "isCurrent": data["isCurrent"],
                    "title": data["title"],
                }
            )
        else:
            await self.ws.send_json(
                {
                    "event": "conversation_deleted",
                    "ok": True,
                    "conversation": conversation,
                    "isCurrent": data["isCurrent"],
                    "title": data["title"],
                }
            )

    async def on_rename_conversation(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket rename conversation"""
        self.logger.info("Renaming conversation: %s", data["conversation"])
        conversation, new_title = data["conversation"], data["new_title"]
        try:
            self.manager.rename_history(conversation, new_title)
        except Exception as ex:
            await self.ws.send_json(
                {
                    "event": "conversation_renamed",
                    "ok": False,
                    "error": str(ex),
                }
            )
        else:
            await self.ws.send_json(
                {
                    "event": "conversation_renamed",
                    "ok": True,
                    "title": new_title,
                    "name": conversation,
                }
            )

    async def on_get_config(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Get profiles and configs"""
        configs = self.config[POOL_KEY]
        profile = data["profile"]

        config = configs[profile]
        default = configs["default"]
        config_schema = {}

        for key, value in CONFIG_SCHEMA.items():
            config_schema[key] = value.copy()
            # Until we support editing the config in the web interface
            config_schema[key]["readonly"] = True
            if "." in key:
                parent, child = key.split(".", 1)
                try:
                    config_schema[key]["value"] = config[parent][child]
                except KeyError:
                    pass
                config_schema[key]["default"] = default[parent][child]
            elif key in config:
                config_schema[key]["value"] = config[key]
                config_schema[key]["default"] = default[key]
            else:
                config_schema[key]["default"] = default[key]

        await self.ws.send_json(
            {
                "event": "config_obtained",
                "configSchema": config_schema,
            }
        )

    # async def on_save_config(self, app: Quart, data: Mapping[str, Any]) -> None:
    #     """Run on websocket save config"""
    #     profile, config_schema = data["profile"], data["configSchema"]
    #     configs = load_toml(MY_CONFIG_FILE)
    #     # New profile
    #     if profile not in configs:
    #         configs[profile] = {}
    #     config = configs[profile]
    #     for key, value in config_schema.items():
    #         if "value" in value:
    #             config[key] = value["value"]

    #     try:
    #         save_toml(MY_CONFIG_FILE, configs)
    #         if profile not in self.config[POOL_KEY]:
    #             self.config[POOL_KEY][profile] = {}
    #         self.config[POOL_KEY][profile].update(config)
    #     except Exception as ex:
    #         await self.ws.send_json(
    #             {
    #                 "event": "config_saved",
    #                 "ok": False,
    #                 "error": str(ex),
    #             }
    #         )
    #     else:
    #         if (profile, "web") in Manager.MANAGERS:
    #             # delete the cache, as the config has changed
    #             # we need to re-init the model
    #             del Manager.MANAGERS[(profile, "web")]

    #         await self.ws.send_json(
    #             {
    #                 "event": "config_saved",
    #                 "ok": True,
    #                 "willLoad": data.get("willLoad", False),
    #                 "profile": profile,
    #             }
    #         )
    #         restart()

    async def on_load_profile(self, app: Quart, data: Mapping[str, Any]) -> None:
        if self.manager:
            self.manager.close()
            self.manager = None

    async def on_delete_profile(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket delete profile"""
        profile = data["profile"]
        with open(MY_CONFIG_FILE) as f:
            configs = yaml.load(f)
        try:
            del configs[profile]
            with open(MY_CONFIG_FILE, "w") as f:
                yaml.dump(configs, f)
            del self.config[POOL_KEY][profile]
        except Exception as ex:
            await self.ws.send_json(
                {
                    "event": "profile_deleted",
                    "ok": False,
                    "error": str(ex),
                }
            )
        else:
            await self.ws.send_json(
                {
                    "event": "profile_deleted",
                    "ok": True,
                }
            )

    async def on_get_docs(self, app: Quart, data: Mapping[str, Any]) -> None:
        """Run on websocket get ingested"""
        p = await asyncio.create_subprocess_exec(
            sys.executable,
            "-m",
            "pxgpt",
            "docs",
            "-p",
            data["profile"],
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
        )
        stdout, stderr = await p.communicate()
        if p.returncode != 0:
            await self.ws.send_json(
                {
                    "event": "docs_failed",
                    "error": stderr.decode(),
                }
            )
        else:
            # parse the stdout
            # It's like
            # Ingested documents:
            # - scGAN.pdf: /path/to/scGAN.pdf
            # Pending documents:
            # - AttentionIsAllYouNeed.pdf: /path/to/AttentionIsAllYouNeed.pdf
            stdout = stdout.decode()
            docs = {"ingested": None, "pending": None}
            for line in stdout.splitlines():
                line = line.strip()
                if not line:
                    continue
                if line == "Ingested documents:":
                    docs["ingested"] = []
                elif line == "Pending documents:":
                    docs["pending"] = []
                elif line.startswith("- "):
                    name, path = line[2:].split(": ", 1)
                    if isinstance(docs["pending"], list):
                        docs["pending"].append({"name": name, "path": path})
                    elif isinstance(docs["ingested"], list):
                        docs["ingested"].append({"name": name, "path": path})

            await self.ws.send_json(
                {
                    "event": "docs_obtained",
                    "docs": docs,
                }
            )



"""Provides a command to solve that async not supported by some chains/llms.

This command will be spawned by the server, by asyncio.create_subprocess_exec().
By communicating with the server via stdin/stdout, it can run asynchronously.
Let me know if you have a better idea.

Don't run this command manually, unless for testing purpose.
The output and logs are not optimized for human reading.
"""
from __future__ import annotations

import logging
import sys

from .logger import get_logger
from .manager import Manager
from .rings._callbacks import ZzzCallbackHandler


def from_input(prefix: str, end: str = "@on_input_end") -> str:
    """Read input from stdin with prefix

    Args:
        prefix: The prefix of the input.
        end: The end of the input.

    Returns:
        The input text.
    """
    text = ""
    while True:
        chars = input()
        if chars == end:
            break
        if not chars.startswith(f"{prefix}:"):
            continue

        text += chars[len(prefix) + 1 :] + "\n"

    return text.strip()


def zzz(profile: str):
    """The entrance of the command.

    Args:
        profile: The profile.
    """
    logger = get_logger("INFO")
    logger.removeHandler(logger.handlers[0])
    # Use sys.stdout for logging, so the parent process can read the logs.
    logger.addHandler(logging.StreamHandler(sys.stdout))
    handler = logger.handlers[0]
    # Add level names so parent process can parse
    handler.setFormatter(
        logging.Formatter("@on_log:%(levelname)s:%(message)s")
    )
    manager = Manager(profile)
    logger.warning(
        "This command is used to solve that async not supported by some LLMs. "
        "Don't run this command manually."
    )
    logger.info("You may check out the issues:")
    logger.info("https://github.com/hwchase17/langchain/issues/5210")
    logger.info("https://github.com/nomic-ai/gpt4all/issues/752")

    # Waiting for inputs.
    # If a KeyboardInterrupt is raised, the program will exit.
    while True:
        try:
            conversation = from_input("@on_conversation_input")
            logger.info("Conversation: %s", conversation)

            query = from_input("@on_query_input")
            logger.info("Query: %r", query)

            manager.query(query, conversation, ZzzCallbackHandler())

        except (EOFError, KeyboardInterrupt):
            logger.info("Exiting...")
            break
        except Exception as exc:
            logger.exception(exc)
            break



"""Serves the web interface"""
from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Awaitable, Callable, Coroutine

from hypercorn.config import Config as HyperConfig
from hypercorn.asyncio import serve as qserve
from quart import Quart, redirect, websocket

from .config import config
from .logger import get_logger
from .ws import WS

if TYPE_CHECKING:
    from logging import Logger


class MyQuart(Quart):
    """Subclass of Quart that uses our logger"""

    def __init__(self, *args, logger: Logger, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._logger = logger

    def run_task(
        self,
        host: str = "127.0.0.1",
        port: int = 5000,
        debug: bool | None = None,
        ca_certs: str | None = None,
        certfile: str | None = None,
        keyfile: str | None = None,
        shutdown_trigger: Callable[..., Awaitable[None]] | None = None,
    ) -> Coroutine[None, None, None]:
        """Return a task that when awaited runs this application.

        This is best used for development only, see Hypercorn for
        production servers.

        Arguments:
            host: Hostname to listen on. By default this is loopback
                only, use 0.0.0.0 to have the server listen externally.
            port: Port number to listen on.
            debug: If set enable (or disable) debug mode and debug output.
            ca_certs: Path to the SSL CA certificate file.
            certfile: Path to the SSL certificate file.
            keyfile: Path to the SSL key file.

        """
        config = HyperConfig()
        config.access_log_format = "%(h)s %(r)s %(s)s %(b)s %(D)s"
        config.accesslog = self.logger
        config.bind = [f"{host}:{port}"]
        config.ca_certs = ca_certs
        config.certfile = certfile
        if debug is not None:
            self.debug = debug
        config.errorlog = config.accesslog
        config.keyfile = keyfile

        return qserve(self, config, shutdown_trigger=shutdown_trigger)


def serve(port):
    """Serve the application with Hypercorn"""
    # Only log_level from default profile works
    logger = get_logger(config.log_level)

    app = MyQuart(
        __package__,
        static_folder=Path(__file__).parent / "frontend" / "build",
        static_url_path="/",
        logger=logger,
    )
    ws = WS()

    @app.route("/", methods=["GET"])
    def _index():
        return redirect(f"index.html")

    @app.websocket("/ws")
    async def _ws():
        """The websocket handler"""
        try:
            while True:
                message = await websocket.receive()
                message = json.loads(message)
                if message["event"] == "connect":
                    ws.ws = websocket._get_current_object()
                    await ws.on_connect(app, message)
                else:
                    fun = getattr(ws, f'on_{message["event"]}')
                    await fun(app, message)
        finally:
            await ws.on_disconnect(app)

    app.run(
        host="0.0.0.0",
        port=port,
        use_reloader=True,
    )



from __future__ import annotations

from typing import List

from langchain.llms.base import LLM
from langchain.callbacks.manager import CallbackManagerForLLMRun


class EchoLLM(LLM):
    """An LLM that echoes the prompt.

    It is used when you don't want to condense your question with chat history.
    Helpful for small models.
    """

    @property
    def _llm_type(self) -> str:
        """Return the LLM type."""
        return "echo"

    def _call(
        self,
        prompt: str,
        stop: List[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
    ) -> str:
        """Echo the prompt."""
        if stop is not None:
            raise ValueError("EchoLLM does not support stop.")
        return prompt



from __future__ import annotations

import logging
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from logging import Logger


def get_logger(log_level: str) -> Logger:
    """Create or get the logger with the given configuration"""
    logger = logging.getLogger("pxgpt")
    logger.setLevel(log_level.upper())
    if logger.hasHandlers():
        return logger

    formatter = logging.Formatter(
        "[%(asctime)s] %(levelname)s %(message)s"
    )

    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    return logger



"""Provides ring for OpenAI models, such as text-davinci-***."""
from __future__ import annotations

from typing import TYPE_CHECKING, Mapping
from langchain.embeddings import OpenAIEmbeddings
from langchain.llms import OpenAI
from langchain.llms.base import LLM

from .base import BaseRing

if TYPE_CHECKING:
    from langchain.embeddings.base import Embeddings


class OpenAIRing(BaseRing):

    MODEL = OpenAI

    @property
    def credentials(self) -> Mapping[str, str]:
        return {
            key: val
            for key, val in self.config.credentials.items()
            if key in (
                "openai_api_key",
                "openai_api_base",
                "openai_organization",
                "openai_proxy",
            )
        }

    @property
    def embeddings(self) -> Embeddings:
        return OpenAIEmbeddings(**self.credentials, **self.config.embeddings)

    def _create_llm(self) -> LLM:
        """Create the LLM object, pass the credentials to it"""
        self.logger.info("Creating LLM (%s)", self.config.model.type)

        return self.MODEL(**self.model_args, **self.credentials)

    def _create_qllm(self) -> LLM | None:
        """Create the QLLM object, pass the credentials to it"""
        if not self.config.ingest.source_directory:
            return None

        qmodel_name = (
            self.config.model.type
            if self.qmodel is None
            else self.qmodel.__name__
        )
        self.logger.info("Creating QLLM (%s)", qmodel_name)
        if self.qmodel is None:
            return self.llm.copy(
                include=set(self.llm.__fields__),
                update={
                    "streaming": False,
                    "verbose": False,
                    "callbacks": None,
                    "tags": ["qmodel"],
                    "metadata": {},
                },
            )

        if "OpenAI" in qmodel_name:
            return self.qmodel(**self.qmodel_args, **self.credentials)

        return self.qmodel(**self.qmodel_args)



"""Provides the CLI chat interface"""
from __future__ import annotations

import asyncio
import time
from typing import List, Tuple

from prompt_toolkit.shortcuts import PromptSession, set_title
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from .version import __version__
from .manager import Manager

from .rings._callbacks import DocumentStreamingStdOutCallbackHandler


CHAT_COMMANDS = {
    "/help": "List the commands",
    "/new": "Start a new conversation",
    "/switch": "Switch to a conversation",
    "/list": "List all conversations",
    "/path": "Show the path of the current conversation file",
    "/delete": "Delete a conversation",
    "/rename": "Rename a conversation",
    "/ingest": "Ingest documents from the source directory",
    "/docs": "List ingested and uningested documents in the source directory",
    "/exit": "Exit the CLI",
}

CHAT_COMMAND_HELPS = {
    "/help": [
        "/help: List all commands",
        "/help [command]: Show the help for a command",
        "/help [/command]: Show the help for a command",
    ],
    "/new": [
        "/new: Start a new conversation",
        "/new [title]: Start a new conversation with a title",
    ],
    "/switch": [
        "/switch [datetime]: Switch to a conversation named like 2021-01-01_00-00-00",
        "/switch [index]: Switch to a conversation by index (use /list to see the index)",
    ],
    "/delete": [
        "/delete: Delete current conversation",
        "/delete [datetime]: Delete a conversation named like 2021-01-01_00-00-00",
        "/delete [index]: Delete a conversation by index (use /list to see the index)",
    ],
    "/rename": [
        "/rename <title>: Rename current conversation to <title>",
    ],
    "/ingest": [
        "/ingest: Ingest documents from the source directory",
        "/ingest force: Force ingestion, even if the document exists",
        "/ingest f: Force ingestion, even if the document exists",
        "/ingest --force: Force ingestion, even if the document exists",
        "/ingest -f: Force ingestion, even if the document exists",
    ],
}


class ChatCommand:
    """Manage the chat commands"""

    def __init__(
        self,
        manager: Manager,
        session: PromptSession,
        history: str,
        history_list: List[str],
    ):
        self.manager = manager
        self.session = session
        self.history = history
        self.history_list = history_list

    @classmethod
    def get_command_and_args(cls, query: str) -> Tuple[str, List[str]] | bool:
        """Get the command and the arguments from a query

        Args:
            query: The query string

        Returns:
            False if query does not start with "/"
            (command, args) if query starts with "/"
            When command is not a known command, returns (command, False)
        """
        if not query.startswith("/"):
            # not a command, continue query
            return False

        cmd, *args = query.split(" ")
        if cmd not in CHAT_COMMANDS:
            # is a command, but not a known one, don't continue query
            return cmd, False

        return cmd, args

    async def a_run(self, cmd: str, args: List[str]) -> bool | None:
        """Run a command asynchronously

        Args:
            cmd: The command
            args: The arguments

        Returns:
            False to break the loop
        """
        method = f"_{cmd[1:]}"
        amethod = f"_a_{cmd[1:]}"

        if hasattr(self, amethod):
            fun = getattr(self, amethod)
            return await fun(args)

        if hasattr(self, method):
            fun = getattr(self, method)
            return fun(args)

        print(f"Error: Unknown command: {cmd}")
        return False

    def run(self, cmd: str, args: List[str]) -> bool | None:
        """Run a command

        Args:
            cmd: The command
            args: The arguments

        Returns:
            False to break the loop
        """
        method = f"_{cmd[1:]}"

        if hasattr(self, method):
            fun = getattr(self, method)
            return fun(args)

        print(f"Error: Unknown command: {cmd}")
        return False

    def _help(self, args: List[str]) -> None:
        """Show the help for a command"""
        if not args:
            print("Commands:")
            for cmd, desc in CHAT_COMMANDS.items():
                print(f"  - {cmd}: {desc}")
        elif len(args) > 1:
            print("Error: Too many arguments")
        else:
            cmd = args[0]
            if not cmd.startswith("/"):
                cmd = "/" + cmd
            if cmd not in CHAT_COMMANDS:
                print(f"Error: Unknown command: {cmd}")
                return
            print(f"Command: {cmd}")
            help = CHAT_COMMAND_HELPS.get(cmd, f"{cmd}: {CHAT_COMMANDS[cmd]}")
            if isinstance(help, str):
                print(f"  - {help}")
            else:
                for h in help:
                    print(f"  - {h}")

    def _list(self, args: List[str]) -> None:
        """List all conversations"""
        print("Conversations:")
        for i, history in enumerate(self.history_list):
            bullet = "*" if history == self.history else "-"
            title = self.manager.history_title(
                history,
                fallback=True,
                add_date=True,
            )
            print(f"  {bullet} {i + 1}. {title}")

    def _new_(self, args: List[str], asyn: bool) -> None:
        """Start a new conversation"""
        title = " ".join(args) if args else None
        history = self.manager.new_history(title)
        self.history_list.insert(0, history)
        # recycle previous session
        del self.session
        _run_history(history, "Created a new conversation: ", self.manager, self.history_list, asyn)

    def _new(self, args: List[str]) -> None:
        """Start a new conversation"""
        self._new_(args, False)
        # exit previous session loop
        return False

    async def _a_new(self, args: List[str]) -> None:
        """Start a new conversation asynchronously"""
        self._new_(args, True)
        return False

    def _switch_(self, args: List[str], asyn: bool) -> None:
        """Switch to a conversation"""
        if not args:
            print("Error: Missing argument")
            return
        if len(args) > 1:
            print("Error: Too many arguments")
            return
        arg = args[0]
        if arg.isdigit():
            try:
                history = self.history_list[int(arg) - 1]
            except IndexError:
                print(f"Error: Invalid index: {arg}")
                return
        else:
            history = arg
            if history not in self.history_list:
                print(f"Error: Unknown conversation: {history}")
                return
        # recycle previous session
        del self.session
        _run_history(history, "Switched to conversation: ", self.manager, self.history_list, asyn)

    def _switch(self, args: List[str]) -> None:
        """Switch to a conversation"""
        self._switch_(args, False)
        # exit previous session loop
        return False

    async def _a_switch(self, args: List[str]) -> None:
        """Switch to a conversation asynchronously"""
        self._switch_(args, True)
        return False

    def _delete_(self, args: List[str], asyn: bool) -> None:
        """Delete a conversation"""
        if len(args) > 1:
            print("Error: Too many arguments")
            return

        arg = args[0]
        if arg.isdigit():
            try:
                history = self.history_list[int(arg) - 1]
            except IndexError:
                print(f"Error: Invalid index: {arg}")
                return
        else:
            history = arg
            if history not in self.history_list:
                print(f"Error: Unknown conversation: {history}")
                return

        self.manager.history_file(history).unlink()
        self.history_list.remove(history)
        print(f"Deleted conversation: {history}")

        if history == self.history:
            if self.history_list:
                del self.session
                # recycle previous session
                _run_history(self.history_list[0], "Switched to conversation: ", self.manager, self.history_list, asyn)
            else:
                self._new_([], asyn)

            return False

    def _delete(self, args: List[str]) -> None:
        """Delete a conversation"""
        return self._delete_(args, False)

    async def _a_delete(self, args: List[str]) -> None:
        """Delete a conversation asynchronously"""
        return self._delete_(args, True)

    def _path(self, args: List[str]) -> None:
        """Show the path of the current conversation file"""
        print("Path of the current conversation file:")
        print(" ", self.manager.history_file(self.history))

    def _rename(self, args: List[str]) -> None:
        """Rename a conversation"""
        if not args:
            print("Error: Missing argument")
            return

        title = " ".join(args)
        self.manager.history_change_title(self.history, title=title)
        print(f"Renamed conversation to: {title}")

    def _ingest(self, args: List[str]) -> None:
        """Ingest documents from the source directory"""
        if self.manager.config.ingest.source_directory is None:
            print("Error: No source directory set in the config")
            return

        force = args and args[0] in ("force", "f", "--force", "-f")
        self.manager.ingest(force=force)

    def _docs(self, args: List[str]) -> None:
        """List ingested and uningested documents in the source directory"""
        if self.manager.config.ingest.source_directory is None:
            print("Error: No source directory set in the config")
            return

        docs = self.manager.get_docs()
        print("Ingested documents:")
        if docs["ingested"]:
            for doc in docs["ingested"]:
                print(f"  - {doc['name']}")
        else:
            print("  - None")
        print()

        print("Documents pending ingestion:")
        if docs["pending"]:
            for doc in docs["pending"]:
                print(f"  - {doc['name']}")

            # print("  You can ingest them with /ingest")
        else:
            print("  - None")

    def _exit(self, args: List[str]) -> None:
        """Exit the CLI"""
        raise EOFError()


async def _a_chat(
    manager: Manager,
    session: PromptSession,
    history: str,
    history_list: List[str],
) -> None:
    """Run the chat asynchronously"""
    last_query_took = 0
    chat_command = ChatCommand(manager, session, history, history_list)
    while True:
        try:
            if last_query_took > 0.0:
                query = await session.prompt_async(
                    bottom_toolbar=f"Last query took {last_query_took} s."
                )
            else:
                query = await session.prompt_async()

            query = query.strip()
            cmd_args = ChatCommand.get_command_and_args(query)
            if cmd_args is not False:
                cmd, args = cmd_args
                if args is False:
                    print(f"Error: Unknown command: {cmd}")
                    continue

                if await chat_command.a_run(cmd, args) is False:
                    break
                else:
                    continue

            start = time.perf_counter()
            await manager.a_query(
                query,
                history,
                callbacks=[DocumentStreamingStdOutCallbackHandler()],
            )

            last_query_took = round(time.perf_counter() - start, 2)
            print()

        except (EOFError, KeyboardInterrupt):
            print("Bye!")
            return


def _chat(
    manager: Manager,
    session: PromptSession,
    history: str,
    history_list: List[str],
) -> None:
    """Run the chat"""
    last_query_took = 0
    chat_command = ChatCommand(manager, session, history, history_list)
    while True:
        try:
            if last_query_took > 0.0:
                query = session.prompt(
                    bottom_toolbar=f"Last query took {last_query_took} s."
                )
            else:
                query = session.prompt()

            query = query.strip()
            if not query:
                continue

            cmd_args = ChatCommand.get_command_and_args(query)
            if cmd_args is not False:
                cmd, args = cmd_args
                if args is False:
                    print(f"Error: Unknown command: {cmd}")
                    continue

                if chat_command.run(cmd, args) is False:
                    break
                else:
                    continue

            start = time.perf_counter()
            manager.query(
                query,
                history,
                callbacks=[DocumentStreamingStdOutCallbackHandler()],
            )

            last_query_took = round(time.perf_counter() - start, 2)
            print()

        except (EOFError, KeyboardInterrupt):
            print("Bye!")
            return


def _run_history(
    history_name: str,
    msg: str,
    manager: Manager,
    history_list: List[str],
    asyn: bool,
) -> None:
    """Run the history/conversation"""

    history_title = manager.history_title(
        history_name,
        fallback=True,
        add_date=True,
    )
    print(msg, history_title)

    history = manager.history_to_prompt_history(history_name)

    print("Use the up/down arrow keys to navigate history.")
    print("Use /help to see the list of commands.")

    session = PromptSession(
        "\n>>> ",
        history=InMemoryHistory(history),
        auto_suggest=AutoSuggestFromHistory(),
    )
    set_title("pxGPT CLI")

    if asyn:
        asyncio.run(_a_chat(manager, session, history_name, history_list))
    else:
        _chat(manager, session, history_name, history_list)


def chat_cli(profile: str, asyn: bool) -> None:
    """Chat with the model in the terminal"""
    print()
    print(f"Welcome to chat via the pxGPT CLI v{__version__}!")
    print("Hit 'Ctrl+c' or 'Ctrl+d' to exit.")

    manager = Manager(profile)
    history = None
    history_list = manager.get_histories()

    if not history_list:
        history_list = [manager.new_history()]
        msg = "Created a new conversation: "
    else:
        msg = "Loaded conversation: "

    history = history_list[0]
    _run_history(history, msg, manager, history_list, asyn)



from __future__ import annotations

from abc import ABC
from functools import cached_property
from multiprocessing import Pool
from pathlib import Path
from typing import TYPE_CHECKING, List, Mapping

from chromadb.config import Settings
from langchain.document_loaders import (
    CSVLoader,
    EverNoteLoader,
    PyMuPDFLoader,
    TextLoader,
    UnstructuredEmailLoader,
    UnstructuredEPubLoader,
    UnstructuredFileLoader,
    UnstructuredHTMLLoader,
    UnstructuredMarkdownLoader,
    UnstructuredODTLoader,
    UnstructuredPowerPointLoader,
    UnstructuredWordDocumentLoader,
    UnstructuredURLLoader,
)
from langchain.vectorstores import Chroma, SKLearnVectorStore
from langchain.text_splitter import RecursiveCharacterTextSplitter

if TYPE_CHECKING:
    from langchain.embeddings.base import Embeddings
    from langchain.docstore.document import Document


class MyElmLoader(UnstructuredEmailLoader):
    """Wrapper to fallback to text/plain when default does not work"""

    def load(self) -> List[Document]:
        """Wrapper adding fallback for elm without html"""
        try:
            try:
                doc = UnstructuredEmailLoader.load(self)
            except ValueError as e:
                if "text/html content not found in email" in str(e):
                    # Try plain text
                    self.unstructured_kwargs["content_source"] = "text/plain"
                    doc = UnstructuredEmailLoader.load(self)
                else:
                    raise
        except Exception as e:
            # Add file_path to exception message
            raise type(e)(f"{self.file_path}: {e}") from e

        return doc


class MyURLLoader(UnstructuredFileLoader):
    """Load the content with the URL in a file"""

    def load(self) -> List[Document]:
        urls = Path(self.file_path).read_text().strip().splitlines()
        loader = UnstructuredURLLoader(
            urls,
            headers={"User-Agent": "Mozilla/5.0", "ssl_verify": False},
        )
        return loader.load()


LOADER_MAPPING = {
    ".csv": (CSVLoader, {}),
    ".doc": (UnstructuredWordDocumentLoader, {}),
    ".docx": (UnstructuredWordDocumentLoader, {}),
    ".enex": (EverNoteLoader, {}),
    ".eml": (MyElmLoader, {}),
    ".epub": (UnstructuredEPubLoader, {}),
    ".html": (UnstructuredHTMLLoader, {}),
    ".md": (UnstructuredMarkdownLoader, {}),
    ".odt": (UnstructuredODTLoader, {}),
    ".pdf": (PyMuPDFLoader, {}),
    ".ppt": (UnstructuredPowerPointLoader, {}),
    ".pptx": (UnstructuredPowerPointLoader, {}),
    ".txt": (TextLoader, {"encoding": "utf8"}),
    ".url": (MyURLLoader, {}),
    ".urls": (MyURLLoader, {}),
}


def load_single_document(file_path: str) -> List[Document]:
    ext = "." + file_path.rsplit(".", 1)[-1]
    if ext in LOADER_MAPPING:
        loader_class, loader_args = LOADER_MAPPING[ext]
        loader = loader_class(file_path, **loader_args)
        return loader.load()

    raise ValueError(f"Unsupported file extension '{ext}'")


class ChromaIngestorMixin(ABC):
    """Ingest the documents into the database.

    Part of the code is borrowed from
    https://github.com/imartinez/privateGPT/blob/main/ingest.py
    by Ivan Martinez.

    Check the LICENSE file for the license of the original code.

    There is a bug with chroma, that each time that it may generte a different
    uuid, so that an empty vector will be returned for the persisted document.
    See: https://github.com/hwchase17/langchain/issues/3011

    So we use SKLearnVectorStore instead.
    """

    @cached_property
    def persist_directory(self) -> str:
        """Get the persist directory"""
        source_directory = Path(self.config.ingest.source_directory)
        if not source_directory.exists():
            raise ValueError(
                f"Source directory {source_directory} does not exist"
            )

        if self.config.ingest.persist_directory is None:
            pd = Path(self.config.ingest.source_directory).joinpath(
                f".pxgpt-{self.config.model.type.lower()}-db"
            )
            pd.mkdir(exist_ok=True)
            pd = str(pd)
        else:
            pd = self.config.ingest.persist_directory

        return pd

    @property
    def embeddings(self) -> Embeddings:
        """The embeddings"""
        raise NotImplementedError

    def db(self, texts: List[Document] | None = None) -> Chroma:
        """Get the database"""
        settings = Settings(
            chroma_db_impl=self.config.ingest.chroma_db_impl,
            persist_directory=self.persist_directory,
            anonymized_telemetry=self.config.ingest.chroma_anonymized_telemetry,
        )
        if texts is None:
            return Chroma(
                persist_directory=self.persist_directory,
                embedding_function=self.embeddings,
                client_settings=settings,
            )

        return Chroma.from_documents(
            texts,
            persist_directory=self.persist_directory,
            embedding=self.embeddings,
            client_settings=settings,
        )

    def get_docs(self) -> Mapping[str, List[Mapping[str, str]]]:
        """Get the ingested files"""
        out = {"ingested": [], "pending": []}
        collection = self.db().get()

        ingested = [
            Path(metadata["source"]).resolve()
            for metadata in collection["metadatas"]
        ]
        persist_directory = Path(self.persist_directory).resolve()
        source_directory = Path(self.config.ingest.source_directory).resolve()
        for doc in source_directory.rglob("*"):
            if (
                doc.name.startswith(".")
                or doc.is_dir()
                or doc.is_relative_to(persist_directory)
                or doc.suffix in (".parquet", ".bin", ".pkl")
            ):
                continue
            if doc in ingested:
                out["ingested"].append({"name": doc.name, "path": str(doc)})
            else:
                out["pending"].append({"name": doc.name, "path": str(doc)})

        return out

    def load_documents(self, ignored_files: List[str]) -> List[Document]:
        """Loads all documents from the source documents directory, ignoring
        specified files
        """
        status = []
        all_files = set()
        source_dir = Path(self.config.ingest.source_directory)

        for ext in LOADER_MAPPING:
            all_files = all_files | set(
                (str(p) for p in source_dir.rglob(f"*{ext}"))
            )

        filtered_files = all_files - set(ignored_files)

        if not status:
            status = [0, len(filtered_files)]  # loaded, total

        with Pool(processes=self.config.ingest.n_workers or None) as pool:
            results = []
            for docs in pool.imap_unordered(
                load_single_document, filtered_files
            ):
                results.extend(list(docs))
                status[0] += 1
                self.logger.info("Loaded %s/%s documents", *status)

        return results

    def process_documents(
        self,
        ignored_files: List[str] | None = None,
    ) -> List[Document]:
        """Load documents and split in chunks"""
        ignored_files = ignored_files or []

        self.logger.info(
            "Loading documents from %s",
            self.config.ingest.source_directory,
        )
        documents = self.load_documents(ignored_files)
        if not documents:
            self.logger.warning("No new documents to load")
            return []

        self.logger.info("Loaded %s new documents", len(documents))
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=self.config.ingest.chunk_size,
            chunk_overlap=self.config.ingest.chunk_overlap,
        )
        texts = text_splitter.split_documents(documents)
        self.logger.info(
            "Split into {%s} chunks of text (max. {%s} tokens each)",
            len(texts),
            self.config.ingest.chunk_size,
        )
        return texts

    def does_vectorstore_exist(self) -> bool:
        """Checks if vectorstore exists"""
        persist_directory = Path(self.persist_directory)

        if (
            not persist_directory.joinpath("index").exists()
            or not persist_directory.joinpath(
                "chroma-collections.parquet"
            ).exists()
            or not persist_directory.joinpath(
                "chroma-embeddings.parquet"
            ).exists()
        ):
            return False

        list_index_files = list(
            persist_directory.joinpath("index").glob("*.bin")
        )
        list_index_files.extend(
            persist_directory.joinpath("index").glob("*.pkl")
        )
        # At least 3 documents are needed in a working vectorstore
        return len(list_index_files) > 3

    def ingest(self, force: bool = False) -> None:
        if self.does_vectorstore_exist():
            # Update and store locally vectorstore
            self.logger.info(
                "Appending to existing vectorstore at %s",
                self.persist_directory,
            )
            db = self.db()
            if force:
                self.logger.info(
                    "Clear current collection for force ingestion"
                )
                db._collection.delete()

            collection = db.get()

            texts = self.process_documents(
                # Ignore already ingested files
                [metadata["source"] for metadata in collection["metadatas"]],
            )

            if not texts:
                return

            self.logger.info("Creating embeddings. May take some minutes...")
            db.add_documents(texts)
        else:
            # Create and store locally vectorstore
            self.logger.info(
                "Creating new vectorstore at %s",
                self.persist_directory,
            )
            texts = self.process_documents()

            if not texts:
                return

            self.logger.info("Creating embeddings. May take some minutes...")
            db = self.db(texts)

        db.persist()
        db = None

        self.logger.info("Ingestion complete!")


class SKLearnInvestorMixin(ChromaIngestorMixin, ABC):

    @property
    def persist_path(self) -> str:
        return str(Path(self.persist_directory).joinpath("db.parquet"))

    def db(self, texts: List[Document] | None = None) -> SKLearnVectorStore:
        """Get the database"""

        if texts:
            vs = SKLearnVectorStore.from_documents(
                documents=texts,
                persist_path=self.persist_path,
                embedding=self.embeddings,
                serializer="parquet",
            )
        else:
            vs = SKLearnVectorStore(
                embedding=self.embeddings,
                persist_path=self.persist_path,
                serializer="parquet",
            )

        return vs

    def get_docs(self) -> Mapping[str, List[Mapping[str, str]]]:
        """Get the ingested files"""
        out = {"ingested": [], "pending": []}
        db = self.db()

        ingested = [
            Path(metadata["source"]).resolve()
            for metadata in db._metadatas
        ]
        persist_directory = Path(self.persist_directory).resolve()
        source_directory = Path(self.config.ingest.source_directory).resolve()
        for doc in source_directory.rglob("*"):
            if (
                doc.name.startswith(".")
                or doc.is_dir()
                or doc.is_relative_to(persist_directory)
                or doc.suffix in (".parquet", ".bin", ".pkl")
            ):
                continue
            if doc in ingested:
                out["ingested"].append({"name": doc.name, "path": str(doc)})
            else:
                out["pending"].append({"name": doc.name, "path": str(doc)})

        return out

    def does_vectorstore_exist(self) -> bool:
        """Checks if vectorstore exists"""
        persist_path = Path(self.persist_path)
        return persist_path.exists()

    def ingest(self, force: bool = False) -> None:
        if self.does_vectorstore_exist():
            # Update and store locally vectorstore
            self.logger.info(
                "Appending to existing vectorstore at %s",
                self.persist_path,
            )
            if force:
                Path(self.persist_path).unlink()

            db = self.db()

            texts = self.process_documents(
                # Ignore already ingested files
                [metadata["source"] for metadata in db._metadatas],
            )

            if not texts:
                return

            self.logger.info("Creating embeddings. May take some minutes...")
            db.add_documents(texts)
        else:
            # Create and store locally vectorstore
            self.logger.info(
                "Creating new vectorstore at %s",
                self.persist_directory,
            )
            texts = self.process_documents()

            if not texts:
                return

            self.logger.info("Creating embeddings. May take some minutes...")
            db = self.db(texts)

        db.persist()
        db = None

        self.logger.info("Ingestion complete!")



from pathlib import Path

from diot import Diot
from simpleconf import ProfileConfig

DEFAULT_CONFIG = Diot(
    log_level="INFO",
    history_directory="~/.config/pxgpt/history",
    history_into_memory=True,
    credentials=Diot(
        openai_api_key=None,
    ),
    embeddings=Diot(),
    model=Diot(
        type="GPT4All",
        model="models/gpt4all/ggml-gpt4all-j-v1.3-groovy.bin",
    ),
    ingest=Diot(
        chunk_size=500,
        chunk_overlap=50,
        source_directory=None,
        n_workers=0,
        persist_directory=None,
        target_source_chunks=4,
    ),
)

CONFIG_SCHEMA = {
    "log_level": Diot(
        default=DEFAULT_CONFIG.log_level,
        kind="choice",
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help=(
            "Log level for the logger in the terminal. Note that the server "
            "needs to be restarted for the change to take effect."
        ),
    ),
    "ingest.source_directory": Diot(
        default=None,
        kind="str",
        help=(
            "The directory to store the source files. "
            "If not set, then a chat chain is used, "
            "instead of a retrieval QA chain."
        ),
    ),
    "ingest.chunk_size": Diot(
        default=500,
        kind="int",
        help="Maximum size of chunks while ingesting the source documents",
    ),
    "ingest.chunk_overlap": Diot(
        default=50,
        kind="int",
        help="Overlap between chunks while ingesting the source documents",
    ),
    "ingest.n_workers": Diot(
        default=0,
        kind="int",
        help=(
            "Number of workers to use while ingesting the source documents. "
            "0 to use all available CPUs"
        ),
    ),
    "ingest.persist_directory": Diot(
        default=None,
        kind="str",
        help=(
            "The directory to store the vectorstore. If not set, "
            "`.pxgpt-<model>-db` under the source_directory is created and used"
        )
    ),
    "credentials.openai_api_key": Diot(
        default=None,
        kind="password",
        placeholder="sk-xxxxxxxxxxxxxxxxxxxxxxxx",
        help="The API key for OpenAI API.",
    ),
    "model.type": Diot(
        default="GPT4All",
        kind="choice",
        choices=["GPT4All", "LlamaCpp", "OpenAI", "ChatOpenAI"],
        help=(
            "The type of the language model to use. "
            "When you choose GPT4All, you can also set `model_allow_download` "
            "to True to download the model from the internet automatically. "
            "When using openai models, if you are not sure about the model type"
            " you can just use `OpenAI` and the model will be automatically "
            "detected."
        ),
    ),
    # "model.<more>": Diot(
    #     default=None,
    #     kind="more",
    #     help=(
    #         "Model specific parameters passed to the language model. "
    #         "Not that `verbose`, `callbacks` and `streaming` are not allowed. "
    #         "You can use `model.allow_download` to download the model from the "
    #         "internet automatically for GPT4All, for example."
    #     ),
    # ),
    # model=Diot(
    #     default="models/gpt4all/ggml-gpt4all-j-v1.3-groovy.bin",
    #     kind="str",
    #     help=(
    #         "The path to the model file or the name of the model. "
    #         "If you allow download for some models, you can specify an "
    #         "unexisting path. The directories will be created automatically. "
    #     ),
    # ),
    "ingest.target_source_chunks": Diot(
        default=4,
        kind="int",
        help=(
            "The number of sources to use for each search. "
        ),
    ),
    "history_directory": Diot(
        default="~/.config/pxgpt/history",
        kind="str",
        help="The directory to store the history files",
    ),
    "history_into_memory": Diot(
        default=True,
        kind="bool",
        help=(
            "Whether to load the history files into memory. "
            "You can turn it off for small models."
        ),
    ),
}

# Used to save modified configs
MY_CONFIG_FILE = Path("./.pxgpt.config.yml")

CONFIG_FILES = [
    {"default": DEFAULT_CONFIG},
    Path("~/.config/pxgpt/config.yml").expanduser(),
    Path("~/.pxgpt.config.yml").expanduser(),
    MY_CONFIG_FILE,
]


config = ProfileConfig.load(*CONFIG_FILES, ignore_nonexist=True)



"""Provides the entrance of the ingest command."""

from .manager import Manager


def ingest_cli(profile: str, force: bool) -> None:
    """The entrance of the command.

    Args:
        profile: The profile.
    """
    manager = Manager(profile)
    if not manager.config.ingest.source_directory:
        print("No source directory specified, nothing to ingest.")
        return

    manager.ingest(force=force)



import argx
from simpleconf import ProfileConfig


def main():
    """The main function"""
    parser = argx.ArgumentParser()

    profile_names = ["--profile", "-p"]
    profile_args = {
        "dest": "profile",
        "help": "Name of the profile to use from the config file",
        "default": "default",
    }

    # ingest
    parser_ingest = parser.add_command(
        "ingest", help="Ingest documents into the database"
    )
    parser_ingest.add_argument(*profile_names, **profile_args)
    parser_ingest.add_argument(
        "--force",
        "-f",
        action="store_true",
        default=False,
        help=(
            "Force ingestion, even if the document exists. "
            "You may want to use this if you have updated the documents or "
            "configurations for ingestion."
        ),
    )
    # chat
    parser_chat = parser.add_command(
        "chat", help="Chat with the model in the terminal"
    )
    parser_chat.add_argument(
        "--asyn",
        action="store_true",
        help=(
            "Use async calls to the model? "
            "Don't use it if you don't know what it is."
        ),
        default=False,
    )
    parser_chat.add_argument(*profile_names, **profile_args)
    # docs
    parser_docs = parser.add_command(
        "docs", help="Show ingested and pending documents under a profile"
    )
    parser_docs.add_argument(*profile_names, **profile_args)
    # serve
    parser_serve = parser.add_command(
        "serve", help="Serve a web interface to chat with the model"
    )
    parser_serve.add_argument(
        "--port",
        "-p",
        help="Port to serve the web interface on",
        default=7758,
        type=int,
    )
    # profiles
    parser.add_command(
        "profiles", help="List available profiles"
    )
    # zzz
    parser_zzz = parser.add_command(
        "zzz",
        help=(
            "A command used to solve that async not supported by some LLMs. "
            "Don't run this command manually."
        )
    )
    parser_zzz.add_argument(*profile_names, **profile_args)

    args = parser.parse_args()

    if args.COMMAND == 'ingest':
        from .ingest import ingest_cli
        ingest_cli(args.profile, args.force)

    elif args.COMMAND == 'chat':
        from .chat import chat_cli
        chat_cli(args.profile, args.asyn)

    elif args.COMMAND == 'serve':
        from .serve import serve
        serve(args.port)

    elif args.COMMAND == 'profiles':
        from .config import config
        profiles = ProfileConfig.profiles(config)
        print("\nAvailable profiles:\n")
        for profile in profiles:
            print(f"- {profile}")
        print()

    elif args.COMMAND == 'docs':
        from .docs import docs_cli
        docs_cli(args.profile)

    elif args.COMMAND == 'zzz':
        from .zzz import zzz
        zzz(args.profile)



"""Provides the Manager class"""
from __future__ import annotations

import time
import json

from functools import lru_cache
from pathlib import Path
from typing import Any, List, Mapping
from langchain.callbacks.base import BaseCallbackHandler

from diot import FrozenDiot
from simpleconf import ProfileConfig

from .config import config
from .logger import get_logger
from .rings.base import NoAsyncSupportRing


class Manager:
    """The ring manager

    Manages the rings and histories.
    """

    def __init__(self, profile: str, serving: bool = False):
        """Initialize the manager

        Args:
            profile: The profile
            serving: Whether the manager is used for serving
                Primarily used for determining whether to initialize the LLMs
                for those rings that do not support async. Since we don't need
                to initialize the LLMs, as they will be initialized by the
                spawned process.
        """
        old_profile = ProfileConfig.current_profile(config)
        if old_profile != profile:
            ProfileConfig.use_profile(config, profile)

        self.config = FrozenDiot(ProfileConfig.detach(config))
        self.profile = profile
        self.logger = get_logger(self.config.log_level)
        if old_profile != profile:
            self.logger.info("Switched profile to %s", profile)

        self.ring = self._create_ring(serving)

    def close(self):
        """Close the manager"""
        del self.ring

    def _create_ring(self, serving: bool):
        """Create the ring"""
        # Avoid changes to the config
        model_type = self.config.model.type

        def should_init_llm(ring_class):
            return not serving or not issubclass(
                ring_class, NoAsyncSupportRing
            )

        if model_type == "GPT4All":
            from .rings.gpt4all import GPT4AllRing

            return GPT4AllRing(
                self.config,
                self.profile,
                init_llm=should_init_llm(GPT4AllRing),
            )
        if model_type == "LlamaCpp":
            from .rings.llamacpp import LlamaCppRing

            return LlamaCppRing(
                self.config,
                self.profile,
                init_llm=should_init_llm(LlamaCppRing),
            )
        if model_type == "OpenAI":
            from .rings.openai import OpenAIRing

            return OpenAIRing(
                self.config, self.profile, init_llm=should_init_llm(OpenAIRing)
            )
        if model_type == "ChatOpenAI":
            from .rings.chatopenai import ChatOpenAIRing

            return ChatOpenAIRing(
                self.config,
                self.profile,
                init_llm=should_init_llm(ChatOpenAIRing),
            )
        else:
            raise ValueError(f"Unknown ring: {model_type}")

    def get_docs(self) -> Mapping[str, List[Mapping[str, str]]]:
        """Get the documents"""
        return self.ring.get_docs()

    def ingest(self, force: bool = False) -> None:
        """Ingest the documents"""
        self.ring.ingest(force=force)

    def get_histories(
        self, include_title: bool = False
    ) -> List[str | Mapping[str, Any]]:
        """Get the list of histories

        Args:
            include_title: Whether to include the title of the history
                If True, each element of the output list will be a dict
                with the keys `name` and `title`. Otherwise, it will be
                a string with the name of the history.

        Returns:
            List[str | Mapping[str, Any]]: The list of histories
        """
        history_dir = (
            Path(self.config.history_directory)
            .joinpath(self.profile)
            .expanduser()
        )

        return [
            hfile.stem
            if not include_title
            else {
                "name": hfile.stem,
                "title": self.history_title(
                    hfile.stem, fallback=True, add_date=False
                ),
            }
            for hfile in sorted(
                history_dir.glob("*.json"),
                reverse=True,
                key=lambda x: x.stat().st_ctime,
            )
        ]

    @lru_cache()
    def history_file(self, history: str, title: str | None = None) -> Path:
        """Get the path to a history file

        If the history file does not exist, it will be created.

        Args:
            history: The history (name) to get the path of
            title: The title of new history. Only used when the history
                file does not exist.

        Returns:
            The path to the history file
        """
        history_dir = (
            Path(self.config.history_directory)
            .joinpath(self.profile)
            .expanduser()
        )
        history_dir.mkdir(parents=True, exist_ok=True)
        out = history_dir.joinpath(f"{history}.json")
        if not out.exists():
            title = json.dumps(title)
            out.write_text(f'{{"title": {title}, "history": []}}\n')
        return out

    def new_history(self, title: str | None = None) -> str:
        """Create a new history, and return its name (stem)

        Args:
            title: The title of the new history

        Returns:
            The name (stem) of the new history
        """
        hf = self.history_file(f"{time.strftime('%Y-%m-%d_%H-%M-%S')}", title)
        self.logger.info("Created new history: %s", hf.stem)
        return hf.stem

    def load_history(self, history: str) -> Mapping[str, Any]:
        """Load a history

        Args:
            history: The history (name) to load

        Returns:
            The history dict
        """
        with self.history_file(history).open() as f:
            return json.load(f)

    def rename_history(self, history: str, new_title: str) -> None:
        """Rename a history

        Args:
            history: The history (name) to rename
            new_title: The new title of the history
        """
        history_file = self.history_file(history)
        with history_file.open() as f:
            history_json = json.load(f)

        history_json["title"] = new_title

        with history_file.open("w") as f:
            json.dump(history_json, f, indent=2)

    def history_title(
        self,
        history: str,
        fallback: bool,
        add_date: bool,
    ) -> str | None:
        """Get the title of a history

        Args:
            history (str): The history (name) to get the title of
            fallback (bool): Whether to fallback to the history name
                if the title is not found or empty. Otherwise, return None.
            add_date (bool): Whether to add the date to the title
                if the title is found and not empty.

        Returns:
            str | None: The title of the history
        """
        with self.history_file(history).open() as f:
            history_json = json.load(f)

        title = history_json.get("title", None)
        if not title and not fallback:
            return None

        if not title:
            return history

        if add_date:
            title = f"{title} ({history})"

        return title

    def history_change_title(self, history: str, title: str) -> None:
        """Change the title of a history

        Args:
            history: The history (name) to change the title of
            title: The new title
        """
        with self.history_file(history).open() as f:
            history_json = json.load(f)

        history_json["title"] = title

        with self.history_file(history).open("w") as f:
            json.dump(history_json, f, indent=2)

    def history_to_prompt_history(self, history: str) -> List[str]:
        """Convert the history to a prompt-toolkit history

        Args:
            history: The history (name) to convert

        Returns:
            The prompt-toolkit history
        """
        with self.history_file(history).open() as f:
            history = json.load(f)

        if "history" not in history:
            return []

        return [
            message["data"]["content"]
            for message in history["history"]
            if message["type"] == "human"
        ]

    def query(
        self,
        query: str,
        history: str,
        callbacks: List[BaseCallbackHandler] | BaseCallbackHandler,
    ) -> Mapping[str, Any]:
        """Query the model with a string"""
        return self.ring.query(query, self.history_file(history), callbacks)

    async def a_query(
        self,
        query: str,
        history: str,
        callbacks: List[BaseCallbackHandler] | BaseCallbackHandler,
    ) -> Mapping[str, Any]:
        """Query the model with a string asynchronously"""
        return await self.ring.a_query(
            query, self.history_file(history), callbacks
        )



"""Provides ring for the LlamaCpp model."""
from __future__ import annotations

from typing import TYPE_CHECKING
from langchain.embeddings import LlamaCppEmbeddings
from langchain.llms import LlamaCpp

from .base import NoAsyncSupportRing

if TYPE_CHECKING:
    from langchain.embeddings.base import Embeddings


class LlamaCppRing(NoAsyncSupportRing):

    MODEL = LlamaCpp

    @property
    def embeddings(self) -> Embeddings:
        return LlamaCppEmbeddings(**self.config.embeddings)

    # def _create_llm(self) -> LlamaCpp:
    #     """Create the LLM object"""
    #     llm = super()._create_llm()
    #     # https://github.com/ggerganov/llama.cpp/issues/999
    #     llm.client.verbose = False
    #     return llm



"""Provides some special callbacks for streaming and other purposes."""

from __future__ import annotations

import sys
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Dict, List
from uuid import UUID, uuid4

from langchain.callbacks.base import AsyncCallbackHandler, BaseCallbackHandler
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.schema import LLMResult

if TYPE_CHECKING:
    from langchain.schema.output import LLMResult


class DocumentStreamingStdOutCallbackHandler(StreamingStdOutCallbackHandler):
    """A callback handler inherited from the StreamingStdOutCallbackHandler
    but prints the source documents.

    This callback handler can be called by both a real chain or called pseudoly
    after parsing the output from the spawned process.
    """

    def on_llm_new_token(
        self, token: str, tags: List[str], **kwargs: Any
    ) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        # Ignore the qmodel, which is used for condense the query.
        if "qmodel" in tags:
            return

        return super().on_llm_new_token(token, **kwargs)

    def on_chain_end(
        self, outputs: Dict[str, Any] | str, **kwargs: Any
    ) -> None:
        """Run when chain ends running."""
        if isinstance(outputs, str) and outputs.startswith(
            "@source_documents:"
        ):
            # from pseduo call
            sources = outputs[len("@source_documents:") :].split(";")
            sys.stdout.write("\n\nSources:")
            for source in sources:
                sys.stdout.write(f"\n- {source}")
            sys.stdout.flush()

        elif "source_documents" in outputs:
            # from real chain call
            sys.stdout.write("\n\nSources:")
            sources = set(
                [doc.metadata["source"] for doc in outputs["source_documents"]]
            )
            for source in sources:
                sys.stdout.write(f"\n- {source}")
            sys.stdout.flush()


class ZzzCallbackHandler(BaseCallbackHandler):
    """A callback handler for the zzz command.

    It prints to stdout, and then the message will be parsed by the parent
    process, and sent to the real callback handlers.
    """

    def _send(
        self, func: str, msg: str | None = None, tags: List[str] = []
    ) -> None:
        """Send message to stdout

        Args:
            func: The callback handler function name
            msg: The message
            tags: The tags of the chain
        """
        msg = msg or ""
        msg = msg.replace("\n", "\\n")
        tag = tags[0] if tags else ""
        sys.stdout.write(f"@{func}:{tag}:{msg}\n")
        sys.stdout.flush()

    def on_llm_new_token(self, token: str, **kwargs) -> Any:
        """Run on new LLM token. Only available when streaming is enabled."""
        self._send("on_llm_new_token", token, tags=kwargs.get("tags", []))

    def on_llm_end(self, response: LLMResult, **kwargs) -> Any:
        """Run when LLM ends running."""
        self._send("on_llm_end", tags=kwargs.get("tags", []))

    def on_llm_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> Any:
        """Run when LLM errors."""
        self._send("on_llm_error", str(error), tags=kwargs.get("tags", []))

    def on_tool_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> Any:
        """Run when tool errors."""
        self._send("on_tool_error", str(error), tags=kwargs.get("tags", []))

    def on_chain_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> Any:
        """Run when chain errors."""
        self._send("on_chain_error", str(error), tags=kwargs.get("tags", []))

    def on_retriever_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> Any:
        """Run when retriever errors."""
        self._send(
            "on_retriever_error", str(error), tags=kwargs.get("tags", [])
        )

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs) -> Any:
        """Run when chain ends running."""
        msg = ""
        if "source_documents" in outputs:
            sources = set(
                [doc.metadata["source"] for doc in outputs["source_documents"]]
            )
            msg = ";".join(sources)
            msg = f"@source_documents:{msg}"
        self._send("on_chain_end", msg, tags=kwargs.get("tags", []))


class StreamingWebsocketCallbackHandler(AsyncCallbackHandler):
    """A callback handler for communicating with the websocket"""

    def __init__(self, ws) -> None:
        """Initialize the callback handler and the websocket"""
        super().__init__()
        self.ws = ws

    async def on_llm_new_token(self, token: str, **kwargs: Any) -> None:
        """Run on new LLM token. Only available when streaming is enabled."""
        with suppress(BrokenPipeError):
            await self.ws.send_json(
                {
                    "event": "responding",
                    "message": token,
                }
            )

    async def on_llm_end(self, response: LLMResult, **kwargs) -> None:
        """Run when LLM ends running."""
        await self.ws.send_json(
            {
                "event": "query_finished",
            }
        )

    async def on_chain_end(
        self,
        outputs: Dict[str, Any],
        *,
        run_id: UUID | None = None,
        **kwargs: Any,
    ) -> None:
        """Run when chain ends running."""
        if run_id is None:  # pseudo call
            run_id = uuid4()
        return await super().on_chain_end(outputs, run_id=run_id, **kwargs)

    async def on_llm_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> None:
        """Run when LLM errors."""
        await self.ws.send_json(
            {
                "event": "lc_error",
                "type": "llm",
                "error": str(error),
            }
        )

    async def on_tool_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> None:
        """Run when tool errors."""
        await self.ws.send_json(
            {
                "event": "lc_error",
                "type": "tool",
                "error": str(error),
            }
        )

    async def on_chain_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> None:
        """Run when chain errors."""
        await self.ws.send_json(
            {
                "event": "lc_error",
                "type": "chain",
                "error": str(error),
            }
        )

    async def on_retriever_error(
        self, error: Exception | KeyboardInterrupt, **kwargs
    ) -> None:
        """Run when retriever errors."""
        await self.ws.send_json(
            {
                "event": "lc_error",
                "type": "retriever",
                "error": str(error),
            }
        )



"""Provides the base ring class for different models and chains"""
from __future__ import annotations
import asyncio

import json
import re
import sys
import textwrap
from abc import ABC
from contextlib import contextmanager
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Type

from langchain.chains import (
    ConversationChain,
    ConversationalRetrievalChain,
    LLMChain,
)
from langchain.chains.question_answering import load_qa_chain
from langchain.callbacks.base import BaseCallbackHandler
from langchain.memory import ChatMessageHistory, ConversationBufferMemory
from langchain.schema.messages import messages_from_dict, messages_to_dict
from langchain.prompts.prompt import PromptTemplate

from ._llms import EchoLLM
from ._mixins import SKLearnInvestorMixin
from ..logger import get_logger

if TYPE_CHECKING:
    from diot import Diot
    from langchain.llms.base import LLM
    from langchain.chains.base import Chain


class DocumentRetrievalBufferMemory(ConversationBufferMemory):
    """Make ConversationBufferMemory work with return_source_documents=True
    See https://github.com/hwchase17/langchain/issues/5630
    """

    def save_context(
        self, inputs: Dict[str, Any], outputs: Dict[str, str]
    ) -> None:
        return super().save_context(inputs, {"response": outputs["answer"]})


class BaseRing(SKLearnInvestorMixin, ABC):
    """Rings are classes that wrap around LLM and Chains from different
    langchain, based on different model types.

    It assembles the LLM and Chain objects, and provides a unified interface
    for the query, with or without documents.

    Args:
        config: A Diot object that contains all the configuration
        logger: A logger object
        init_llm: Whether to initialize the LLMs in the constructor
    """

    MODEL: LLM | None = None

    def __init__(
        self,
        config: Diot,
        profile: str,
        init_llm: bool = True,
    ) -> None:
        """Initialize the ring"""
        self.config = config
        self.profile = profile
        self.logger = get_logger(config.log_level)

        if init_llm:
            self.llm = self._create_llm()
            self.qllm = self._create_qllm()
        else:
            self.llm = None
            self.qllm = None

        if self.config.ingest.source_directory:
            self.logger.info("Entering document retrieval mode")
        else:
            self.logger.info("Entering chat mode")

    @property
    def model_args(self):
        """Get the model arg that will be passed to the model"""
        args = {k: v for k, v in self.config.model.items() if k != "type"}
        args.update({"streaming": True, "verbose": False})
        return args

    @property
    def qmodel(self) -> Type[LLM]:
        """Get the qmodel class

        When qmodel is not found in config, then class of model will be used.
        """
        if "qmodel" not in self.config:
            return None

        qm = self.config.qmodel.get("type", self.config.model.type)
        if qm == self.config.model.type:
            return None

        if qm == "Echo":
            return EchoLLM
        if qm == "OpenAI":
            from langchain.llms import OpenAI

            return OpenAI
        if qm == "ChatOpenAI":
            from langchain.chat_models import ChatOpenAI

            return ChatOpenAI
        if qm == "GPT4All":
            from langchain.llms import GPT4All

            return GPT4All
        if qm == "LlamaCpp":
            from langchain.llms import LlamaCpp

            return LlamaCpp

        raise ValueError(f"Unsupported qmodel type '{qm}'")

    @property
    def qmodel_args(self):
        """Get the qmodel arg that will be passed to the question condenser"""
        qargs = self.config.get("qmodel", {}).copy()
        qargs = {k: v for k, v in qargs.items() if k != "type"}
        qargs.update(
            {"streaming": False, "verbose": False, "tags": ["qmodel"]}
        )
        return qargs

    def _create_llm(self) -> LLM:
        """Create the LLM object

        Custom rings should override this method if MODEL is not set
        """
        self.logger.info("Creating LLM (%s)", self.config.model.type)
        if self.MODEL is None:
            raise NotImplementedError

        return self.MODEL(**self.model_args)

    def _create_qllm(self) -> LLM | None:
        """Create the LLM for question condensing

        When config.ingest.source_directory is not set, this method will not be
        called.

        When self.qmodel is None, self.llm is copied, with some fields updated,
        including steaming=False, verbose=False, callbacks=None,
        tags=["qmodel"] and metadata={}.
        """
        if not self.config.ingest.source_directory:
            return None

        qmodel_name = (
            self.config.model.type
            if self.qmodel is None
            else self.qmodel.__name__
        )
        self.logger.info("Creating QLLM (%s)", qmodel_name)
        if self.qmodel is None:
            return self.llm.copy(
                include=set(self.llm.__fields__),
                update={
                    "streaming": False,
                    "verbose": False,
                    "callbacks": None,
                    "tags": ["qmodel"],
                    "metadata": {},
                },
            )

        return self.qmodel(**self.qmodel_args)

    def _create_chat_chain(self) -> Chain:
        """Create the chat chain object for chatting only, without documents

        A ConversationChain object is created.
        Custom rings can override this method to create a different chain.
        """
        self.logger.debug("Creating chat chain")
        prompt_template = textwrap.dedent(
            """
            The following is a friendly conversation between a human and an AI.
            The AI is talkative and provides lots of specific details from its
            context. If the AI does not know the answer to a question,
            it truthfully says it does not know. After answering a question,
            the AI should NOT ask follow-up questions, followed by the "Human:"
            prefix.

            Previous conversation:
            {history}

            Current conversation:
            Human: {input}
            AI:"""
        ).lstrip()
        prompt = PromptTemplate.from_template(prompt_template)
        return ConversationChain(llm=self.llm, prompt=prompt)

    def _create_retrieval_chain(self) -> Chain:
        """Create the retrieval chain object for document retrieval

        A ConversationalRetrievalChain object is created.
        Custom rings can override this method to create a different chain.
        If you want to use the default chain, you can choose to use a
        different question condenser, or even not to use one. Then you must
        be aware that the "qmodel" section in the config will be ignored.
        """
        self.logger.debug("Creating retrieval chain")
        main_chain = load_qa_chain(self.llm, verbose=False)
        cqp = textwrap.dedent(
            """
            Given the following conversation and a follow up question,
            rephrase the follow up question to be a standalone question,
            in its original language. The new question should be
            grammatically correct and semantically similar to the original
            question. If the chat history is not enough, you can keep the
            original question as is.

            Chat History:
            {chat_history}

            Follow Up Input: {question}
            Standalone question:"""
        ).lstrip()
        condense_question_chain = LLMChain(
            llm=self.qllm,
            prompt=PromptTemplate.from_template(cqp),
            verbose=False,
            tags=["qchain"],
        )
        return ConversationalRetrievalChain(
            retriever=self.db().as_retriever(
                search_kwargs={"k": self.config.ingest.target_source_chunks}
            ),
            combine_docs_chain=main_chain,
            question_generator=condense_question_chain,
            return_source_documents=True,
            memory=DocumentRetrievalBufferMemory(memory_key="chat_history"),
            get_chat_history=lambda x: x,
        )

    @cached_property
    def chain(self):
        """The chain object for chatting or document retrieval"""
        if self.config.ingest.source_directory:
            return self._create_retrieval_chain()
        return self._create_chat_chain()

    @contextmanager
    def _history_context(self, history: Path):
        """A context manager that loads and saves history

        With it, when subclasses override the _chat_chain_call and
        _retrieval_chain_call methods, they don't need to worry about
        loading and saving history.

        Args:
            history: The history file path
        """
        with history.open() as f:
            data = json.load(f)

        memory = self.chain.memory
        if self.config.history_into_memory:
            messages = messages_from_dict(data.get("history", []))
            memory.chat_memory = ChatMessageHistory(messages=messages)

        if not self.chain.metadata:
            self.chain.metadata = {}

        yield

        if self.config.history_into_memory:
            data["history"] = messages_to_dict(memory.chat_memory.messages)
        else:
            data["history"].extend(
                messages_to_dict(memory.chat_memory.messages)
            )

        with history.open("w") as f:
            json.dump(data, f, indent=2)

    def _chat_chain_call(
        self,
        query: str,
        callbacks: List[BaseCallbackHandler],
    ) -> Mapping[str, Any]:
        """Query the model with a string

        Note that if you have different input keys for you chain, you should
        override this method.
        """
        self.logger.debug("Chat calling: %r", query)
        return self.chain(
            {"input": query},
            callbacks=callbacks,
            return_only_outputs=True,
        )

    async def _a_chat_chain_call(
        self,
        query: str,
        callbacks: List[BaseCallbackHandler],
    ) -> Mapping[str, Any]:
        """Query the model with a string asynchronously

        Note that if you have different input keys for you chain, you should
        override this method.
        """
        self.logger.debug("Chat calling asynchronously: %r", query)
        return await self.chain.acall(
            {"input": query},
            callbacks=callbacks,
            return_only_outputs=True,
        )

    def _retrieval_chain_call(
        self,
        query: str,
        callbacks: List[BaseCallbackHandler],
    ) -> Mapping[str, Any]:
        """Query the model with a string

        Note that if you have different input keys for you chain, you should
        override this method.
        """
        self.logger.debug("Retrieval calling: %r", query)
        return self.chain(
            {"question": query, "chat_history": []},
            callbacks=callbacks,
            return_only_outputs=True,
        )

    async def _a_retrieval_chain_call(
        self,
        query: str,
        callbacks: List[BaseCallbackHandler],
    ) -> Mapping[str, Any]:
        """Query the model with a string asynchronously

        Note that if you have different input keys for you chain, you should
        override this method.
        """
        self.logger.debug("Retrieval calling asynchronously: %r", query)
        return await self.chain.acall(
            {"question": query, "chat_history": []},
            callbacks=callbacks,
            return_only_outputs=True,
        )

    def query(
        self,
        query: str,
        history: Path,
        callbacks: List[BaseCallbackHandler] | BaseCallbackHandler,
    ) -> Mapping[str, Any]:
        """Query the model with a string"""
        if isinstance(callbacks, BaseCallbackHandler):
            callbacks = [callbacks]

        self.logger.debug("Querying: %r", query)
        with self._history_context(history):
            if self.config.ingest.source_directory:
                return self._retrieval_chain_call(query, callbacks)
            return self._chat_chain_call(query, callbacks)

    async def a_query(
        self,
        query: str,
        history: Path,
        callbacks: List[BaseCallbackHandler] | BaseCallbackHandler,
    ) -> Mapping[str, Any]:
        """Query the model with a string asynchronously"""
        if isinstance(callbacks, BaseCallbackHandler):
            callbacks = [callbacks]

        self.logger.debug("Querying asynchronously: %r", query)
        with self._history_context(history):
            if self.config.ingest.source_directory:
                return await self._a_retrieval_chain_call(query, callbacks)
            return await self._a_chat_chain_call(query, callbacks)


async def _pseduo_callbacks_call(
    callbacks: List[BaseCallbackHandler],
    func: str,
    *args: Any,
    **kwargs: Any,
) -> None:
    """Call the passed-in callback handlers in a pseudo way

    This is used in the workaround for async not supported by some chains/llms.
    The information are gathered from stdout of the spawned process, so
    most of the information are lost. So you pass in the callbacks, pay
    attention that if they use some information that are not supported.

    Args:
        callbacks: The callback handlers, could be either sync or async
        func: The function name to call
        args: The args to pass to the function
        kwargs: The kwargs to pass to the function
    """
    for cb in callbacks:
        called = getattr(cb, func)(*args, **kwargs)
        # if called is a coroutine, await it
        if hasattr(called, "__await__"):
            await called


class NoAsyncSupportRing(BaseRing, ABC):
    """A ring with a workaround for async not supported by some chains/llms.

    Not that with the workaround, we are not able to get the final results,
    everything has to be done with the callbaks.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the ring"""
        super().__init__(*args, **kwargs)
        self.spawned = None

    def __del__(self):
        """Terminate the spawned process"""
        if self.spawned:
            self.logger.debug("Terminating spawned process")
            self.spawned.terminate()
            self.spawned = None

    async def _a_call(
        self,
        query: str,
        history: Path,
        callbacks: List[BaseCallbackHandler],
    ) -> None:
        """Query the model in a spawned process calling the zzz subcommand

        Args:
            query: The query
            history: The history file path
            callbacks: The callback handlers
        """
        if not self.spawned:
            self.logger.debug("Spawning process")
            self.spawned = await asyncio.create_subprocess_exec(
                sys.executable,
                "-m",
                "pxgpt",
                "zzz",
                "--profile",
                self.profile,
                stdin=asyncio.subprocess.PIPE,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )

        conversation = history.stem
        # pass the conversation
        self.spawned.stdin.write(
            f"@on_conversation_input:{conversation}\n".encode()
        )
        self.spawned.stdin.write(b"@on_input_end\n")
        # pass the query
        for line in query.splitlines():
            self.spawned.stdin.write(f"@on_query_input:{line}\n".encode())
        self.spawned.stdin.write(b"@on_input_end\n")
        await self.spawned.stdin.drain()

        # parse the output, like on_llm_new_token::token
        regex = re.compile(r"^@on_([\w_]+):([\w_]*):(.*)")
        async for line in self.spawned.stdout:
            line = line.decode().rstrip("\n")
            match = regex.match(line)
            if not match:
                # ignore unmatched lines
                continue

            func, tag, msg = match.groups()
            msg = msg.replace("\\n", "\n")

            # logging messages
            if func == "log":
                getattr(self.logger, tag.lower())(f"[spawned] {msg}")
                continue

            tags = [tag] if tag else []
            func = f"on_{func}"
            await _pseduo_callbacks_call(callbacks, func, msg, tags=tags)
            # main chain (without tags) ends
            if func == "on_chain_end" and not tag:
                break

    async def a_query(
        self,
        query: str,
        history: Path,
        callbacks: List[BaseCallbackHandler] | BaseCallbackHandler,
    ) -> Mapping[str, Any]:
        """Query the model with a string asynchronously"""
        if isinstance(callbacks, BaseCallbackHandler):
            callbacks = [callbacks]

        self.logger.debug("Querying with async workaround: %r", query)
        # History will be saved by subprocess
        await self._a_call(query, history, callbacks)



"""Provides ring for GPT4All."""
from __future__ import annotations

from typing import TYPE_CHECKING
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.llms import GPT4All

from .base import NoAsyncSupportRing

if TYPE_CHECKING:
    from langchain.embeddings.base import Embeddings


class GPT4AllRing(NoAsyncSupportRing):

    MODEL = GPT4All

    @property
    def embeddings(self) -> Embeddings:
        return HuggingFaceEmbeddings(**self.config.embeddings)



"""Provides ring for ChatOpenAI models, such as gpt-3.5-turbo"""
from langchain.chat_models import ChatOpenAI

from .openai import OpenAIRing


class ChatOpenAIRing(OpenAIRing):

    MODEL = ChatOpenAI



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pxgpt', 'pxgpt.rings']

package_data = \
{'': ['*'], 'pxgpt': ['frontend/build/*', 'frontend/build/assets/*']}

install_requires = \
['argx>=0.2.9,<0.3.0',
 'chromadb>=0.3.29,<0.4.0',
 'langchain>=0.0.232,<0.0.233',
 'prompt-toolkit>=3.0.39,<4.0.0',
 'python-simpleconf[yaml]>=0.6.0,<0.7.0',
 'quart>=0.18.4,<0.19.0']

extras_require = \
{'all': ['gpt4all>=1.0.3,<2.0.0',
         'llama-cpp-python>=0.1.70,<0.2.0',
         'openai>=0.27.8,<0.28.0'],
 'gpt4all': ['gpt4all>=1.0.3,<2.0.0'],
 'llama-cpp': ['llama-cpp-python>=0.1.70,<0.2.0'],
 'openai': ['openai>=0.27.8,<0.28.0']}

entry_points = \
{'console_scripts': ['pxgpt = pxgpt.cli:main']}

setup_kwargs = {
    'name': 'pxgpt',
    'version': '0.0.0rc0',
    'description': 'Your personal, powerful and private GPT.',
    'long_description': '# <img src="pxgpt/frontend/public/assets/favicon.png" alt="p" width=28 />GPT: Your personal, powerful and private GPT\n\n## Features\n\n- Ingest of your own documents and talk to them.\n- Store your data locally on your device.\n- Choose from a variety of models, including OpenAI.\n- Support conversation history and memory.\n- Switch between profiles with different settings.\n- Support both web interface and command line interface.\n\n## Installation\n\n```shell\n# With all supported models\n$ pip install -U pxgpt[all]\n\n# With support for GPT4all only\n$ pip install -U pxgpt[gpt4all]\n\n# With support for llama-cpp only\n$ pip install -U pxgpt[llama-cpp]\n\n# With support for openai only\n$ pip install -U pxgpt[openai]\n```\n\n## Usage\n\n### Chat from CLI\n\n```shell\n$ 23:45:06 ❯ pxgpt chat\n\nWelcome to chat via the pxgpt CLI v0.0.0!\nHit \'Ctrl+c\' or \'Ctrl+d\' to exit.\n[2023-07-18 23:45:09,969] INFO Creating LLM (GPT4All)\nFound model file at  models/gpt4all/ggml-gpt4all-j-v1.3-groovy.bin\ngptj_model_load: loading model from \'models/gpt4all/ggml-gpt4all-j-v1.3-groovy.bin\' - please wait ...\ngptj_model_load: n_vocab = 50400\ngptj_model_load: n_ctx   = 2048\ngptj_model_load: n_embd  = 4096\ngptj_model_load: n_head  = 16\ngptj_model_load: n_layer = 28\ngptj_model_load: n_rot   = 64\ngptj_model_load: f16     = 2\ngptj_model_load: ggml ctx size = 5401.45 MB\ngptj_model_load: kv self size  =  896.00 MB\ngptj_model_load: ................................... done\ngptj_model_load: model size =  3609.38 MB / num tensors = 285\n[2023-07-18 23:45:11,741] INFO Entering chat mode\nLoaded conversation:  2023-07-16_23-11-27\nUse the up/down arrow keys to navigate history.\nUse /help to see the list of commands.\n\n>>> Hello?\n  Hi there! How may I help you today?\n\n>>> /help\nCommands:\n  - /help: List the commands\n  - /new: Start a new conversation\n  - /switch: Switch to a conversation\n  - /list: List all conversations\n  - /path: Show the path of the current conversation file\n  - /delete: Delete a conversation\n  - /rename: Rename a conversation\n  - /ingest: Ingest documents from the source directory\n  - /docs: List ingested and uningested documents in the source directory\n  - /exit: Exit the CLI\n\n>>>\n```\n\n### Chat from the web interface\n\n```shell\n$ pxgpt serve\n# Open http://localhost:7758 in your browser\n```\n\n![Web-interface](web-interface.png)\n\n### Configuration\n\nThe configuration files are loaded from the following paths:\n\n- `~/.config/pxgpt/config.yml`\n- `~/.pxgpt.config.yml`\n- `./.pxgpt.config.yml`\n\n#### Profiles\n\nNote that you need to define profiles in the configuration file. For example:\n\n```yaml\nopenai:  # The profile\n    model:\n        type: ChatOpenAI\n```\n\nThe configuration items are inherited from the `default` profile. For example:\n\n```yaml\ndefault:\n    credentials:\n        openai_api_key: sk-xxxxxxxxxxx\n\nopenai:\n    model:\n        type: ChatOpenAI\n```\n\nThen when you use `openai` profile, the configurations are expanded as:\n\n```yaml\nopenai:\n    credentials:\n        openai_api_key: sk-xxxxxxxxxxx\n    model:\n        type: ChatOpenAI\n```\n\nHigher-level configurations override lower-level configurations. For example:\n\nIf you define the `default` profile in `~/.config/pxgpt/config.yml` and the `openai` profile in `./.pxgpt.config.yml`, then the `openai` profile will inherit the `default` profile, as well.\n\n#### Configuration items\n\n- `log_level`: The log level for the logger in your teminal\n- `history_directory`: The directory to store the conversation history\n- `history_into_memory`: Whether to load the conversation history into memory\n  - You can turn this off if you are using small models\n- `credentials`: The credentials for the models.\n  For example, for OpenAI, you need to provide the `openai_api_key`.\n- `model`: Type of the model and arguments for it.\n  - `type`: The type of the model, supported models are: `GPT4All`, `LlamaCpp`, `ChatOpenAI` and `OpenAI`\n  - `<other>`: The arguments for the model. Passed to `langchain` llms.\n    - For `GPT4All`, you can pass the arguments listed in [here][1].\n    - For `LlamaCpp`, you can pass the arguments listed in [here][2].\n    - For `ChatOpenAI`, you can pass the arguments listed in [here][3].\n    - For `OpenAI`, you can pass the arguments listed in [here][4].\n- `qmodel`: The arguments for model used to condense questions\n  - `type`: Same as `model.type`, with and `Echo` model added, which is useful for models that don\'t do question condensing very well.\n  - `<other>`: Same as `model.<other>`.\n- `ingest`: The arguments for the ingestion.\n  - `source_directory`: The directory to ingest documents from.\n    - If not provided, we will enter the chat mode.\n  - `persist_directory`: The directory to save the vectorstore database.\n    - If not provided, will use `<source_directory>/.pxgpt-<model>-db`.\n  - `target_source_chunks`: The number of chunks to return against the query.\n  - `n_workers`: The number of workers to use for ingestion.\n  - `chunk_size` and `chunk_overlap`: The chunk size and overlap for the ingestion.\n- `embeddings`: The arguments for the embeddings.\n  - For `GPT4All`, you can pass the arguments listed in [here][5].\n  - For `LlamaCpp`, you can pass the arguments listed in [here][6].\n  - For `OpenAI` or `ChatOpenAI`, you can pass the arguments listed in [here][7].\n\n### Ingest documents\n\n```shell\n$ pxgpt ingest  # default profile\n$ pxgpt ingest --profile openai-docs\n# Will ingest documents under `ingest.source_directory` under `openai-docs` profile\n```\n\n## Credits\n\n`pxgpt` is Inspired by [privateGPT][8], with the addition of openai API support, history and memory support, and a web interface.\n\n## TODO\n\n- [ ] Support ingestion management (upload/download/delete/ingest documents) from the web interface\n- [ ] Support profile management (add/remove/modify) from the web interface\n- [ ] Build a docker image\n- [ ] Support more models\n\n## Q & A\n\n[QA.md](QA.md)\n\n[1]: https://api.python.langchain.com/en/latest/llms/langchain.llms.gpt4all.GPT4All.html#langchain.llms.gpt4all.GPT4All\n[2]: https://api.python.langchain.com/en/latest/llms/langchain.llms.llamacpp.LlamaCpp.html#langchain.llms.llamacpp.LlamaCpp\n[3]: https://api.python.langchain.com/en/latest/chat_models/langchain.chat_models.openai.ChatOpenAI.html#langchain.chat_models.openai.ChatOpenAI\n[4]: https://api.python.langchain.com/en/latest/llms/langchain.llms.openai.OpenAI.html#langchain.llms.openai.OpenAI\n[5]: https://api.python.langchain.com/en/latest/embeddings/langchain.embeddings.huggingface.HuggingFaceEmbeddings.html#langchain.embeddings.huggingface.HuggingFaceEmbeddings\n[6]: https://api.python.langchain.com/en/latest/embeddings/langchain.embeddings.llamacpp.LlamaCppEmbeddings.html#langchain.embeddings.llamacpp.LlamaCppEmbeddings\n[7]: https://api.python.langchain.com/en/latest/embeddings/langchain.embeddings.openai.OpenAIEmbeddings.html#langchain.embeddings.openai.OpenAIEmbeddings\n[8]: https://github.com/imartinez/privateGPT\n',
    'author': 'pwwang',
    'author_email': 'pwwang@pwwang.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""Provides the entrance of the ingest command."""

from .manager import Manager


def docs_cli(profile: str) -> None:
    """The entrance of the command.

    Args:
        profile: The profile.
    """
    manager = Manager(profile, serving=True)
    if not manager.config.ingest.source_directory:
        print("Error: No source directory specified, nothing to show.")
        return

    docs = manager.get_docs()
    print("\nIngested documents:")
    for doc in docs['ingested']:
        print(f"- {doc['name']}: {doc['path']}")

    print("\nPending documents:")
    for doc in docs['pending']:
        print(f"- {doc['name']}: {doc['path']}")
